import io
import json
import math
import time

import control
import numpy as np
import pandas as pd

from typing import List
from pycollimator.api import Api
from pycollimator.diagrams import Block
from pycollimator.error import NotFoundError
from pycollimator.i18n import N
from pycollimator.log import Log
from pycollimator.models import Model, ModelParameters, load_model
from pycollimator.simulation_hashed_file import SimulationHashedFile
from pycollimator.utils import is_uuid


class LinearizationResult:
    def __init__(self, A, B, C, D):
        self.A = A
        self.B = B
        self.C = C
        self.D = D

    def to_state_space(self):
        """
        convert the linearization result to a python control object
        """
        return control.StateSpace(self.A, self.B, self.C, self.D)

    def __repr__(self):
        def tostring(x):
            return np.array2string(x, separator=",").replace("\n", "").replace(" ", "")

        return (
            f"<{self.__class__.__name__} A={tostring(self.A)} B={tostring(self.B)} "
            + f"C={tostring(self.C)} D={tostring(self.D)}>"
        )

    @classmethod
    def _from_csv(cls, csv_text):
        a_mat, b_mat, c_mat, d_mat = cls.__reshape_results_linearization(csv_text)
        return cls(a_mat, b_mat, c_mat, d_mat)

    # FIXME this format is not very robust
    @classmethod
    def __reshape_results_linearization(csv, results_text):
        read_st = -1
        lines = results_text.splitlines()
        for each in lines:
            if read_st == -1:
                if each[:3] == "dim":
                    # do the processing common for all matrices to get their dimensions
                    line_data = each.split(",")
                    mat_name = line_data[1]
                    tmp_dim = np.fromiter((float(x) for x in line_data[2:4]), dtype=int)
                    read_st = 0
                    if mat_name == "A":
                        a_dims = tmp_dim.copy()
                    elif mat_name == "B":
                        b_dims = tmp_dim.copy()
                    elif mat_name == "C":
                        c_dims = tmp_dim.copy()
                    elif mat_name == "D":
                        d_dims = tmp_dim.copy()
                    else:
                        print("unrecognized matrix name")
            else:
                # do the processing common for all matrices tobget their data
                line_data = each.split(",")
                line_data = line_data[:-1]
                # print(line_data)
                tmp = np.fromiter((float(x) for x in line_data), dtype=float)
                read_st = -1
                if mat_name == "A":
                    a_mat = np.reshape(tmp, a_dims)
                elif mat_name == "B":
                    b_mat = np.reshape(tmp, b_dims)
                elif mat_name == "C":
                    c_mat = np.reshape(tmp, c_dims)
                elif mat_name == "D":
                    d_mat = np.reshape(tmp, d_dims)
                else:
                    print("unrecognized matrix name")
        # get column names instead of uuids.
        return a_mat, b_mat, c_mat, d_mat


class SimulationResults:
    def __init__(self, stream, model: Model):
        if isinstance(stream, str):
            stream = io.StringIO(stream)
        self._raw_df = pd.read_csv(stream)
        self._model = model

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} {len(self.to_pandas().index)} rows x {len(self.to_pandas().columns)} columns>"
        )

    def __getitem__(self, item):
        if item == "time":
            return self._raw_df["time"]  # pylint: disable=unsubscriptable-object
        if is_uuid(item):
            blk = self._model.find_block(uuid=item)
        elif isinstance(item, Block):
            blk = self._model.find_block(uuid=item.uuid)
        else:
            try:
                blk = self._model.find_block(name=item, case=True)
            except NotFoundError:
                try:
                    blk = self._model.find_block(name=item, case=False)
                except NotFoundError:
                    blk = self._model.find_block(item)
        df = self.to_pandas(uuid=blk.uuid)
        return df

    def to_pandas(self, pattern=None, name=None, uuid=None, type=None, case=False) -> pd.DataFrame:
        if is_uuid(pattern):
            uuid = pattern
            pattern = None

        blocks = self._model.find_blocks(pattern=pattern, name=name, uuid=uuid, type=type, case=case)

        cols_by_uuid = []
        for block in blocks:
            for col in self._raw_df.columns:  # type: str
                if col.startswith(block.uuid):
                    cols_by_uuid.append(col)

        cols_by_path = []
        for col in cols_by_uuid:
            path = self._model.get_block_path(col.split("[")[0])
            idx = col.split("[")[1].split("]")[0]
            if path is not None:
                if idx == "0":
                    cols_by_path.append(path)
                else:
                    cols_by_path.append(f"{path}[{idx}]")

        df = self._raw_df[cols_by_uuid]  # pylint: disable=unsubscriptable-object
        df.columns = cols_by_path
        df.index = self._raw_df["time"]  # pylint: disable=unsubscriptable-object
        df.index.name = "time"
        return df

    @property
    def columns(self) -> list:
        pd_columns = self.to_pandas().columns.tolist()
        cols = []
        for col in pd_columns:
            uuid = col.split("[")[0]
            block = is_uuid(uuid) and self._model.find_block(uuid=uuid) or None
            if block is not None:
                cols.append(block.path)
            else:
                cols.append(col)
        return cols


class SimulationLogLine:
    def __init__(self, data):
        self._data = data
        self._str: str = ""
        self._parse(data)

    def __str__(self) -> str:
        return self._str

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self._data}>"

    def _parse(self, data):
        try:
            js_data = data
            if isinstance(data, str):
                js_data = json.loads(data)
            ts = ""
            if js_data.get("timestamp") is not None:
                ts = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(js_data["timestamp"])) + " "
            parsed_line = f"{ts}{js_data.get('level')} {js_data.get('message')}"
            for k, v in js_data.items():
                if k not in ["timestamp", "level", "message"]:
                    parsed_line += f" {k}={v}"
            self._str = parsed_line
        except Exception as e:
            Log.trace(f"Failed to parse log line: {e}")
            self._str = str(data)


class Simulation:
    def __init__(self, data: dict, model: Model):
        self.model = model
        self._data = data
        self._logs: List[SimulationLogLine] = None
        self._results: pd.DataFrame = None

    def __getitem__(self, key):
        return self._data[key]

    def __repr__(self) -> str:
        if Log.is_level_above("DEBUG"):
            return f"<{self.__class__.__name__} status='{self.status}' uuid='{self.uuid}'>"
        return f"<{self.__class__.__name__} status='{self.status}'>"

    @property
    def uuid(self) -> str:
        return self._data["uuid"]

    @property
    def status(self) -> str:
        return self._data["status"]

    @property
    def model_uuid(self) -> str:
        return self._data["model_uuid"]

    @property
    def results_available(self) -> bool:
        return self._data.get("results_available") is True

    @property
    def is_failed(self) -> bool:
        return self.status == "failed"

    @property
    def is_completed(self) -> bool:
        return self.status == "completed"

    @property
    def is_running(self) -> bool:
        return self.status != "created" and not self.is_failed and not self.is_completed

    @property
    def is_started(self) -> bool:
        return self.status != "created"

    @property
    def logs(self) -> List[SimulationLogLine]:
        if self._logs is None:
            logs = Api.simulation_logs(self.model_uuid, self.uuid)
            self._logs = [SimulationLogLine(s) for s in str(logs).splitlines()]
        return self._logs

    @property
    def compilation_logs(self) -> SimulationLogLine:
        logs = self._data.get("compilation_logs")
        if logs is None:
            return None
        if isinstance(logs, list):
            return SimulationLogLine(logs[0])
        return SimulationLogLine(logs)

    def show_logs(self) -> None:
        # It would be awesome to use markdown to pretty-print these logs
        if self.is_failed:
            print(self.compilation_logs)
            return
        for log_line in self.logs:
            print(log_line)

    @property
    def results(self) -> SimulationResults:
        if not self.results_available:
            Log.warning("Simulation results may not be available yet:", self)
            # But we will try downloading them anyway since that's what the user asked for
        results = Api.simulation_results(self.model_uuid, self.uuid)
        return SimulationResults(results, self.model)

    def get_results(self, wait=True) -> SimulationResults:
        if wait:
            self.wait()
        return self.results

    def update(self):
        Log.trace("Updating simulation:", self)
        self._data = Api.simulation_get(self.model_uuid, self.uuid)
        return self

    def start(self):
        Log.debug("Starting simulation:", self)
        self._data = Api.simulation_start(self.model_uuid, self.uuid)

    def wait(self):
        if self.is_completed:
            return

        if not self.is_started:
            Log.debug("Simulation not started:", self)
            self.start()

        start_ts = time.time()
        while self.is_running:
            Log.debug(f"Waiting for simulation {math.trunc(time.time() - start_ts)}s:", self)
            self.update()
            if self.status == "completed":
                break
            time.sleep(1)
        Log.debug("Simulation completed:", self)

    def to_pandas(self, wait=True):
        return self.get_results(wait=wait).to_pandas()

    def _upload_hashed_files(self, model, simulation):
        parameter_overrides = []
        blocks = model._get_blocks_with_data()
        for block_uuid in blocks:
            input_data = blocks[block_uuid]
            sdf = SimulationHashedFile(input_data)
            resp = sdf.upload(model.uuid, simulation.uuid)
            Log.trace(f"Uploaded hashed file for block '{block_uuid}':", resp)
            parameter_overrides.append(
                {
                    "expression": f'__hashed_file("{sdf.hash}", "{sdf.content_type}")',
                    "block_uuid": block_uuid,
                    "parameter_name": "file_name",
                }
            )
        Log.trace("Parameter overrides:", parameter_overrides)
        return parameter_overrides


def _run_simulation_common(model, linearization, parameters, wait, no_sync, ignore_cache) -> Simulation:
    Log.trace("run_simulation:", model.__repr__())
    if not isinstance(model, Model):
        model = load_model(model)

    if not no_sync:
        model.sync()

    if linearization is not None:
        if is_uuid(linearization):
            linearization = {"submodel_uuid": linearization}
        elif isinstance(linearization, Block):
            linearization = {"submodel_uuid": linearization.uuid}
        else:
            raise ValueError(N("Invalid argument 'linearization'"))

    # Prepare DataSource files
    hashed_files = model._get_hashed_files()
    overrides = [
        {
            "expression": f'__hashed_file("{hashed_file.hash}", "{hashed_file.content_type}")',
            "block_uuid": block_uuid,
            "parameter_name": "file_name",
        }
        for block_uuid, hashed_file in hashed_files.items()
    ]

    # Send simulation specific parameters
    parameters = {
        **model.parameters.to_api_data(),
        **(ModelParameters(parameters or {}).to_api_data()),
    }

    # Simulation-specific configuration
    configuration = model.configuration.to_dict()

    sim_request = {
        "version": model.version,
        "no_start": True,
        "configuration": configuration,
        "overrides": overrides,
        "parameters": parameters,
        "ignore_cache": ignore_cache,
    }

    if linearization is not None:
        sim_request["linearization"] = linearization

    sim = Simulation(Api.simulation_create(model.uuid, sim_request), model)

    if sim.status == "created":
        # Upload DataSource files and then update the BlockParameterOverrides
        parameter_overrides = sim._upload_hashed_files(model, sim)
        if len(parameter_overrides) > 0:
            body = {"overrides": parameter_overrides}
            sim._data = Api.simulation_parameters_set(model.uuid, sim.uuid, body)
        sim.start()

    if wait:
        sim.wait()

    if sim.is_failed:
        Log.error(N("Simulation failed! Use .show_logs() to see the logs."))

    return sim


def run_simulation(model, parameters: dict = None, wait=True, no_sync=False, ignore_cache=False) -> Simulation:
    """
    Run a simulation on the given model.
    """
    return _run_simulation_common(model, None, parameters, wait, no_sync, ignore_cache)


def linearize(model, submodel, parameters: dict = None, no_sync=False, ignore_cache=False):
    """
    Linearize the submodel within a model.
    """
    if not isinstance(model, Model):
        model = load_model(model)

    if not isinstance(submodel, Block):
        submodel = model.find_block(submodel)

    sim = _run_simulation_common(model, submodel, parameters, True, no_sync, ignore_cache)

    if not sim.is_completed:
        Log.error(N("Simulation could not complete!"))
        sim.show_logs()
        return sim

    lin_results_csv = Api.linearization_results_csv(model.uuid, sim.uuid)
    res = LinearizationResult._from_csv(lin_results_csv)
    return res



from pycollimator.api import Api
from pycollimator.error import NotFoundError
from pycollimator.i18n import N
from pycollimator.log import Log
from pycollimator.utils import is_uuid


class Project:
    def __init__(self, data):
        self._data = data
        # api_dict_safe_copy(self.__dict__, data)

    def __getitem__(self, key):
        return self._data[key]

    def __str__(self) -> str:
        return self["title"]

    def __repr__(self) -> str:
        if Log.is_level_above("DEBUG"):
            return f"<{self.__class__.__name__} title='{self['title']}' uuid='{self['uuid']}'>"
        return f"<{self.__class__.__name__} title='{self['title']}'>"

    @property
    def uuid(self) -> str:
        return self["uuid"]


def get_project(title: str = None, case=False) -> Project:
    """
    Get the current project or another project by name.
    """

    prj = None
    if title is None:
        prj = Api.get_project()
    elif is_uuid(title):
        prj = Api.get_project(title)
    else:
        projects = Api.get_projects()
        for p in projects["projects"]:
            if p.get("title") == title:
                prj = p
                break
        if not case:
            for p in projects["projects"]:
                if p.get("title", "").lower() == title.lower():
                    prj = p
                    break

    if prj is None:
        raise NotFoundError(N(f"Project not found: '{title}'"))

    return Project(prj)



import typing as T
import requests
from uuid import uuid4

from pycollimator.i18n import N
from pycollimator.log import Log
from pycollimator.error import CollimatorAuthenticationError, CollimatorApiError, CollimatorRetryableApiError
from pycollimator.global_variables import GlobalVariables


class Api:
    BASE_PATH = "/api/v0"

    @classmethod
    def _handle_error(cls, response: T.Any, headers: T.Dict[str, str] = None):
        request_id = (headers is not None) and headers.get("X-Request-ID") or None
        # correlation_id = headers and headers.get('X-Correlation-ID') or None
        if response.status_code >= 400:
            if response.status_code == 401:
                Log.debug("API error 401, response:", response.text)
                raise CollimatorAuthenticationError(
                    N("Authentication failed. Please check your token or reload the page.")
                )
            elif response.status_code == 504:
                Log.trace("API error 504, retrying...")
                raise CollimatorRetryableApiError
            else:
                Log.debug("API error, status_code:", response.status_code, "response:", response.text)
                raise CollimatorApiError(
                    f"API call failed (status: {response.status_code}). Support request ID: "
                    + f"{request_id}. Detailed error:\n{response.text}"
                )

    # TODO add streaming support for logs and results
    @classmethod
    def _call_api(
        cls, api, method, body: T.Any = None, headers: dict = None, body_type="json", response_type="json", retries=3
    ):
        try:
            url = GlobalVariables.url() + api
            headers = headers or GlobalVariables.custom_headers()
            headers["X-Collimator-API-Caller"] = "python-notebook"
            headers["X-Request-ID"] = str(uuid4())
            if body is None:
                Log.trace(method, url)
                response = requests.request(method, url, headers=headers)
            elif body_type == "json":
                Log.trace(method, url, body)
                headers["Content-Type"] = "application/json"
                response = requests.request(method, url, json=body, headers=headers)
            elif body_type == "files":
                Log.trace(method, url, body)
                response = requests.request(method, url, files=body, headers=headers)
            cls._handle_error(response, headers)
            Log.trace(method, "response:", response.text)
            if response_type == "text":
                return response.text
            return response.json()
        except CollimatorRetryableApiError as e:
            if retries > 0:
                Log.trace(method, "retrying...")
                return cls._call_api(api, method, body, headers, body_type, response_type, retries - 1)
            raise e

    @classmethod
    def get(cls, api):
        return cls._call_api(api, method="GET")

    @classmethod
    def post(cls, api, body):
        return cls._call_api(api, method="POST", body=body)

    @classmethod
    def put(cls, api, body):
        return cls._call_api(api, method="PUT", body=body)

    @classmethod
    def delete(cls, api, body):
        return cls._call_api(api, method="DELETE", body=body)

    @classmethod
    def get_projects(cls):
        api = f"{Api.BASE_PATH}/projects"
        return cls.get(api)

    @classmethod
    def get_project(cls, project_uuid: str = None) -> dict:
        project_uuid = project_uuid or GlobalVariables.project_uuid()
        api = f"{Api.BASE_PATH}/projects/{project_uuid}"
        return cls.get(api)

    @classmethod
    def get_model(cls, model_uuid: str) -> dict:
        api = f"{Api.BASE_PATH}/models/{model_uuid}"
        return cls.get(api)

    @classmethod
    def simulation_create(cls, model_uuid: str, body: dict) -> dict:
        api = f"{Api.BASE_PATH}/models/{model_uuid}/simulations"
        return cls.post(api, body)

    @classmethod
    def simulation_get(cls, model_uuid: str, simulation_uuid: str) -> dict:
        api = f"{Api.BASE_PATH}/models/{model_uuid}/simulations/{simulation_uuid}"
        return cls.get(api)

    @classmethod
    def simulation_start(cls, model_uuid: str, simulation_uuid: str) -> dict:
        api = f"{Api.BASE_PATH}/models/{model_uuid}/simulations/{simulation_uuid}/start"
        return cls.post(api, None)

    @classmethod
    def simulation_parameters_set(cls, model_uuid: str, simulation_uuid: str, body: dict) -> dict:
        api = f"{Api.BASE_PATH}/models/{model_uuid}/simulations/{simulation_uuid}/parameters"
        return cls.put(api, body)

    @classmethod
    def simulation_update(cls, model_uuid: str, simulation_uuid: str, body: dict) -> dict:
        api = f"{Api.BASE_PATH}/models/{model_uuid}/simulations/{simulation_uuid}"
        return cls.put(api, body)

    @classmethod
    def simulation_hashed_file_create(cls, model_uuid: str, simulation_uuid: str, body: dict) -> dict:
        api = f"{Api.BASE_PATH}/models/{model_uuid}/simulations/{simulation_uuid}/hashed_files"
        return cls.post(api, body)

    @classmethod
    def simulation_hashed_file_upload(cls, path: str, upload_url: str) -> dict:
        with open(path, "rb") as file:
            return cls._call_api(upload_url, method="PUT", body={"file": file}, body_type="files")

    # FIXME should be a stream
    @classmethod
    def simulation_logs(cls, model_uuid: str, simulation_uuid: str) -> str:
        api = f"{Api.BASE_PATH}/models/{model_uuid}/simulations/{simulation_uuid}/logs"
        return cls._call_api(api, method="GET", response_type="text")

    # FIXME should be a stream
    @classmethod
    def simulation_results(cls, model_uuid: str, simulation_uuid: str, retries=3) -> str:

        api = f"{Api.BASE_PATH}/models/{model_uuid}/simulations"
        api += f"/{simulation_uuid}/process_results?files=continuous_results.csv"

        s3_urls_resp = cls._call_api(api, method="GET")
        url = s3_urls_resp["s3_urls"][0]["url"]
        try:
            response = requests.request("GET", url)
            cls._handle_error(response)
            return response.text
        except CollimatorRetryableApiError as e:
            if retries > 0:
                Log.trace("simulation_results", "retrying...")
                return cls.simulation_results(model_uuid, simulation_uuid, retries - 1)
            raise e

    # @classmethod
    # def model_configuration_update(cls, model_uuid: str, body: dict) -> dict:
    #     api = f"{Api.BASE_PATH}/models/{model_uuid}/configuration"
    #     return cls.put(api, body)

    @classmethod
    def linearization_results_csv(cls, model_uuid: str, simulation_uuid: str) -> str:
        api = f"{Api.BASE_PATH}/models/{model_uuid}/simulations/{simulation_uuid}/lin_results"
        return cls._call_api(api, method="GET", response_type="text")



import os
import re
import typing as T
import warnings

from pycollimator.i18n import N
from pycollimator.log import Log
from pycollimator.utils import is_uuid


class GlobalVariables:
    """
    global variables stores information about user's authentication and the project
    folder they are currently in
    """

    _instance: "GlobalVariables" = None
    _envRegex = re.compile("https://(dev|test|app).collimator.ai")

    def __new__(cls):
        if not cls._instance:
            cls._instance = super(GlobalVariables, cls).__new__(cls)
        return cls._instance

    def __init__(self) -> None:
        if os.environ.get("CI") == "true":
            return
        self.set_auth_token(None)

    def set_auth_token(self, token: str, project: str = None, api_url: str = None) -> None:
        self.auth_token = token or self._detect_auth_token()
        self.project = project or self._detect_project()
        self.api_url = api_url or self._detect_api_url()
        Log.trace("auth_token:", self.auth_token, "project:", self.project, "api_url:", self.api_url)

    def _detect_auth_token(self) -> T.Optional[str]:
        if os.environ.get("AUTH_TOKEN") is not None:
            return os.environ["AUTH_TOKEN"]
        if os.path.isfile("token.txt"):
            with open("token.txt", "r") as token_file:
                return token_file.read().strip()
        warnings.warn(N("No token file found. Please call `collimator.set_auth_token()` to authenticate manually."))

    def _detect_project(self) -> str:
        if os.environ.get("PROJECT_UUID") is not None:
            return os.environ["PROJECT_UUID"]
        path = os.path.abspath(os.curdir)
        project_uuid = path.split("/")[-1]
        if not is_uuid(project_uuid):
            warnings.warn(N("Unable to detect the current project. API calls may fail."))
        return project_uuid

    def _detect_api_url(self) -> str:
        endpoint = None
        if os.environ.get("API_ENDPOINT") is not None:
            endpoint = os.environ["API_ENDPOINT"]
        elif os.path.isfile("environment.txt"):
            with open("environment.txt", "r") as environment_file:
                endpoint = environment_file.read().strip()
        if endpoint is not None:
            if endpoint.startswith("http://localhost"):
                return "http://localhost"
            if endpoint.startswith("http://host.docker.internal"):
                return "http://host.docker.internal"
            try:
                return GlobalVariables._envRegex.match(endpoint).group(0)
            except BaseException:
                warnings.warn(N("Unable to detect the environment, assuming https://app.collimator.ai."))
                return "https://app.collimator.ai"
        warnings.warn(N("Unable to detect the environment, assuming https://app.collimator.ai."))
        return "https://app.collimator.ai"

    @classmethod
    def _get_instance(cls) -> "GlobalVariables":
        if cls._instance is None:
            cls()
        return cls._instance

    @classmethod
    def project_uuid(cls):
        """
        stores the project uuid associated with the folder.
        """
        return cls._get_instance().project

    @classmethod
    def token(cls):
        """
        stores the authentication token
        """
        return cls._get_instance().auth_token

    @classmethod
    def url(cls):
        """
        stores the url, used for environment logic
        """
        return cls._get_instance().api_url

    @classmethod
    def custom_headers(cls) -> T.Dict[str, str]:
        """
        stores the authentication headers used in all API requests
        """
        # FIXME I think requests should be handling the content-type header for us
        custom_headers = {
            "X-Collimator-API-Token": cls._get_instance().auth_token,
            "Accept": "application/json",
            # "content-type": "application/json",
        }
        return custom_headers


def get_project_url() -> str:
    return GlobalVariables.url() + "/projects/" + GlobalVariables.project_uuid()


def set_auth_token(token: str, project_uuid: str = None, api_url: str = None) -> None:
    GlobalVariables._get_instance().set_auth_token(token, project_uuid, api_url)



import json
import math


import numpy as np
import pandas as pd
import typing as T

from simpleeval import EvalWithCompoundTypes

from pycollimator.api import Api
from pycollimator.diagrams import ModelGraph
from pycollimator.error import NotFoundError, NotLoadedError
from pycollimator.global_variables import GlobalVariables
from pycollimator.i18n import N
from pycollimator.log import Log
from pycollimator.projects import get_project
from pycollimator.simulation_hashed_file import SimulationHashedFile
from pycollimator.utils import is_uuid

# from collimator.simulations import run_simulation


# TODO pull these defaults from the schemas
# TODO add support for all model configuration parameters
DEFAULT_STOP_TIME = 10.0
DEFAULT_SAMPLE_TIME = 0.1
DEFAULT_INTERPOLATION = 0.01


class ModelConfiguration:
    def __init__(self, data: T.Any, model: "Model"):
        self.model = model
        self._data = data or {}
        self._stop_time = float(self._data.get("stop_time", DEFAULT_STOP_TIME))
        self._discrete_step = float(self._data.get("sample_time", DEFAULT_SAMPLE_TIME))
        self._interpolation = float(self._data.get("continuous_time_result_interval", DEFAULT_INTERPOLATION))
        self._needs_resync = False
        # TODO parse events and other solver settings

    # FIXME use a proper serializer
    def to_dict(self) -> dict:
        return {
            **self._data,
            "stop_time": self.stop_time,
            "sample_time": self.discrete_step,
            "continuous_time_result_interval": self.interpolation,
        }

    def __repr__(self) -> str:
        d = self.to_dict()
        if d.get("__developer_options", None) is not None and not Log.is_level_above("TRACE"):
            del d["__developer_options"]
        return d.__repr__()

    def __getitem__(self, key):
        if key == "stop_time":
            return self.stop_time
        elif key == "discrete_step":
            return self.discrete_step
        elif key == "interpolation":
            return self.interpolation
        else:
            raise KeyError("key must be one of 'stop_time', 'discrete_step', 'interpolation'")

    def __setitem__(self, key, value: T.Any):
        if key == "stop_time":
            self.stop_time = value
        elif key == "discrete_step":
            self.discrete_step = value
        elif key == "interpolation":
            self.interpolation = value
        else:
            raise KeyError("key must be one of 'stop_time', 'discrete_step', 'interpolation'")

    @property
    def stop_time(self) -> float:
        return self._stop_time

    @stop_time.setter
    def stop_time(self, value: float):
        if self._stop_time != value and value > 0:
            self._stop_time = value
            self._needs_resync = True

    @property
    def discrete_step(self) -> float:
        return self._discrete_step

    @discrete_step.setter
    def discrete_step(self, value: float):
        if self._discrete_step != value and value > 0:
            self._discrete_step = value
            self._needs_resync = True

    @property
    def interpolation(self) -> float:
        return self._interpolation

    @interpolation.setter
    def interpolation(self, value: float):
        if self._interpolation != value and value > 0:
            self._interpolation = value
            self._needs_resync = True

    # def sync(self, remote_model_data: dict = None):
    #     if not self._needs_resync:
    #         return
    #     # should we merge with the remote model data?
    #     request_data = {
    #         "configuration": self.to_dict(),
    #     }
    #     resp = Api.model_configuration_update(self.model.uuid, request_data)
    #     self._needs_resync = False
    #     return resp


class ParameterEvaluator:
    @classmethod
    def __for_math_names(cls) -> T.Dict[str, T.Any]:
        for name in dir(math):
            if not name.startswith("_") and not callable(getattr(math, name)):
                yield name, getattr(math, name)

    @classmethod
    def __for_math_functions(cls) -> T.Dict[str, T.Any]:
        for name in dir(math):
            if not name.startswith("_") and callable(getattr(math, name)):
                yield name, getattr(math, name)

    @classmethod
    def __math_names(cls) -> T.Dict[str, T.Any]:
        return {name: value for name, value in cls.__for_math_names()}

    @classmethod
    def __math_functions(cls) -> T.Dict[str, T.Any]:
        return {name: value for name, value in cls.__for_math_functions()}

    @classmethod
    def all_functions(cls) -> T.Dict[str, T.Any]:
        return {**cls.__math_functions()}

    @classmethod
    def all_names(cls) -> T.Dict[str, T.Any]:
        return {**cls.__math_names()}

    @classmethod
    def evaluate(cls, name: str, expression: str) -> T.Any:
        evaluator = EvalWithCompoundTypes(functions=cls.all_functions(), names=cls.all_names())
        try:
            literal = evaluator.eval(expression)
            return literal
        except Exception as e:
            raise ValueError(N(f"Could not evaluate parameter '{name}' with expression '{expression}': {e}"))


# A quick and simple way to print out numpy arrays in a way that cmlc can handle.
# This will not scale much but should get us somewhere.
def _serialize_numpy_array(param: np.ndarray) -> str:
    lst = param.tolist()
    s = json.dumps(lst, ensure_ascii=True, allow_nan=True, indent=None, separators=(",", ":"))
    # return f"np.array({s})"
    return s


class Parameter:
    def __init__(self, name, expression, is_string=False):
        self.name = name
        self.expression = expression
        self.is_string = is_string

    @property
    def value(self) -> T.Any:
        if self.is_string:
            return self.expression
        return ParameterEvaluator.evaluate(self.name, self.expression)

    def __str__(self) -> str:
        if self.is_string:
            return str(self.expression)
        if isinstance(self.expression, np.ndarray):
            return _serialize_numpy_array(self.expression)
        return str(self.expression)

    def __repr__(self) -> str:
        return f"'{self.expression}'"

    def to_api_data(self) -> T.Dict:
        if isinstance(self.expression, np.ndarray):
            return {"value": _serialize_numpy_array(self.expression)}
        elif self.is_string:
            return {"value": self.expression, "is_string": True}
        else:
            return {"value": str(self.expression)}


class ParameterSet:
    def __init__(self, data: dict = None):
        self._data = data or {}
        self._parameters: T.Dict[str, Parameter] = {}
        for name, param in data.items():  # type: str, dict
            if isinstance(param, dict):
                self._parameters[name] = Parameter(name, param.get("value", None), param.get("is_string", False))
            elif isinstance(param, str):
                self._parameters[name] = Parameter(name, param, is_string=True)
            else:
                self._parameters[name] = Parameter(name, param)

    def __repr__(self) -> str:
        return self._parameters.__repr__()

    def __getitem__(self, key):
        return self._parameters[key]

    def __setitem__(self, key, value: T.Any):
        return self.set(key, value)

    def keys(self):
        return self._parameters.keys()

    def items(self):
        return self._parameters.items()

    def values(self):
        return self._parameters.values()

    # def __setattr__(self, key: str, value: T.Any) -> None:
    #     if key.startswith("_"):
    #         super().__setattr__(key, value)
    #     return self.set(key, value)

    def to_api_data(self) -> dict:
        return {name: param.to_api_data() for name, param in self._parameters.items()}

    def get(self, name: str) -> T.Any:
        return self._parameters[name].value

    def set(self, name: str, literal: T.Any):
        self._parameters[name].expression = str(literal)


class ModelParameters(ParameterSet):
    def __init__(self, data: dict):
        super().__init__(data)


class Model:
    """
    Contents of a model, may be partially loaded.
    """

    def __init__(self, data):
        self._data = data
        self._graph: ModelGraph = None
        self._datasource_data_by_uuid: T.Dict[str, pd.DataFrame] = {}  # block_uuid -> data (pd.DataFrame)
        self._hashed_files: T.Dict[str, SimulationHashedFile] = {}  # block_uuid -> SimulationHashedFile
        self._configuration = ModelConfiguration(data.get("configuration", {}), self)
        self._parameters = ModelParameters(data.get("parameters", {}))

    def __getitem__(self, key):
        return self._data[key]

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        if Log.is_level_above("DEBUG"):
            return (
                f"<{self.__class__.__name__} name='{self.name}' uuid='{self.uuid}' "
                + f"version={self.version} is_loaded={self.is_loaded}>"
            )
        return f"<{self.__class__.__name__} name='{self.name}'>"

    @property
    def name(self) -> str:
        return self._data["name"]

    @property
    def uuid(self) -> str:
        return self._data["uuid"]

    @property
    def project_uuid(self) -> str:
        return self._data["project_uuid"]

    @property
    def url(self) -> str:
        return f"{GlobalVariables.url()}/editor/{self.project_uuid}/{self.uuid}"

    @property
    def version(self):
        return self._data.get("version")

    @property
    def configuration(self) -> ModelConfiguration:
        return self._configuration

    @property
    def parameters(self) -> ModelParameters:
        return self._parameters

    @property
    def is_loaded(self) -> bool:
        """
        Returns True if the model is fully loaded.
        """
        return self._graph is not None
        # return self._data.get("diagram") is not None

    def load(self):
        """
        Load the model, if it's not already loaded
        """
        if not self.is_loaded:
            self.reload()

    def reload(self):
        """
        Reload the model, if it's loaded or not
        """
        data = Api.get_model(self.uuid)
        self._data = data
        self._graph = ModelGraph(data, model=self)
        self._configuration = ModelConfiguration(data.get("configuration"), self)
        self._parameters = ModelParameters(data.get("parameters", {}))

    def sync(self):
        """
        Synchronize local changes with the database and back
        """
        data = Api.get_model(self.uuid)
        self._data["version"] = data["version"]
        self._graph = ModelGraph(data, model=self)
        # resp = self.configuration.sync(data)
        # if resp is not None and resp.get("version") is not None:
        #     self._data["version"] = resp["version"]

    def find_block(
        self, pattern: str = None, name: str = None, uuid: str = None, type: str = None, case=False, autoload=True
    ):
        """
        Find a block by name.
        """
        blocks = self.find_blocks(pattern=pattern, name=name, uuid=uuid, type=type, case=case, autoload=autoload)
        if len(blocks) == 0:
            raise NotFoundError(
                N(
                    f"No block was found in model '{self}' "
                    f"(pattern='{pattern}' name='{name}' uuid='{uuid}' type='{type}')"
                )
            )
        if len(blocks) > 1:
            raise NotFoundError(
                N(
                    f"Multiple matching blocks found in model '{self}' "
                    f"(pattern='{pattern}' name='{name}' uuid='{uuid}' type='{type}'): {blocks}"
                )
            )
        return blocks[0]

    def find_blocks(
        self, pattern: str = None, name: str = None, uuid: str = None, type: str = None, case=False, autoload=True
    ):
        """
        Find all blocks that match a name.
        """
        if not self.is_loaded:
            if not autoload:
                raise NotLoadedError(N(f"Model '{self}' is not loaded"))
            self.load()
        return self._graph.find_blocks(pattern=pattern, name=name, uuid=uuid, type=type, case=case)

    def get_block_path(self, block, autoload=True):
        if not self.is_loaded:
            if not autoload:
                raise NotLoadedError(N(f"Model '{self}' is not loaded"))
            self.load()
        return self._graph.get_block_path(block)

    # Circular imports blah blah blah
    # def run(self, parameters: dict = None, wait=True, no_sync=False, ignore_cache=False):
    #     """
    #     Start a simulation.

    #     This is an alias for collimator.run_simulation.
    #     """
    #     return run_simulation(self, parameters=parameters, wait=wait, no_sync=no_sync, ignore_cache=ignore_cache)

    def set_parameters(self, parameters: dict):
        paramset = ParameterSet(parameters)
        for k, v in paramset.items():
            self._parameters[k] = v

    def get_parameters(self) -> ModelParameters:
        return self._parameters

    def set_configuration(self, configuration: dict = None, **kwargs):
        if configuration is not None:
            for k, v in configuration.items():
                self.configuration[k] = v
        for k, v in kwargs.items():
            self.configuration[k] = v

    def get_configuration(self) -> ModelConfiguration:
        return self._configuration

    def _set_datasource_data(self, block, data: pd.DataFrame):
        self._datasource_data_by_uuid[block.uuid] = data
        self._hashed_files.clear()

    def _get_blocks_with_data(self) -> T.Dict[str, pd.DataFrame]:
        """block_uuid -> data (pd.DataFrame)"""
        return self._datasource_data_by_uuid

    def _prepare_hashed_files(self) -> T.Dict[str, SimulationHashedFile]:
        for block_uuid in self._datasource_data_by_uuid:
            Log.trace(f"Preparing hashed files for block '{block_uuid}'")
            block_data = self._datasource_data_by_uuid[block_uuid]
            sdf = SimulationHashedFile(block_data)
            self._hashed_files[block_uuid] = sdf
        return self._hashed_files

    def _get_hashed_files(self) -> T.Dict[str, SimulationHashedFile]:
        if len(self._hashed_files) == 0:
            self._prepare_hashed_files()
        return self._hashed_files


def load_model(name: str, project: str = None, load=True, case=False) -> Model:
    """
    Get a model from the API by name
    """
    if isinstance(name, Model):
        if load:
            name.load()
        return name

    if is_uuid(name):
        Log.debug(f"Getting model by uuid '{name}'")
        return Model(Api.get_model(name))

    prj = get_project(project)
    for m in prj["models"]:  # type: dict
        Log.trace(f"Checking model '{json.dumps(m)}'")
        if m.get("name") == name:
            model = Model(m)
            if load:
                model.load()
            return model

    if not case:
        Log.debug(f"No model named '{name}' found in project '{project}', trying case insentitive search")
        for m in prj["models"]:  # type: dict
            if m.get("name", "").lower() == name.lower():
                model = Model(m)
                if load:
                    model.load()
                return model

    raise NotFoundError(N(f"Model not found: '{name}'"))


def list_models() -> T.List[Model]:
    """
    list all models available to be loaded in to the project
    """
    model_list_json = Api.get_project()
    return [Model(model_list_json["models"][i]) for i in range(len(model_list_json["models"]))]



#!/bin/env python3

"""Setup script for collimator notebook package."""

from setuptools import setup, find_packages  # type: ignore

deps = ["numpy", "pandas", "requests", "aiohttp", "control", "simpleeval"]

setup(
    name="pycollimator",
    setup_requires=['setuptools_scm'],
    use_scm_version={
        'root': '../../../',
        'relative_to': __file__,
        'local_scheme': 'no-local-version',
        'version_scheme': 'post-release',
    },
    packages=find_packages(),
    python_requires=">=3.7",
    install_requires=deps,
    long_description="Python package for Collimator.ai"
)



"""Main package to import to use collimator for notebook execution"""

from pycollimator.api import Api
from pycollimator.global_variables import get_project_url, set_auth_token
from pycollimator.log import Log
from pycollimator.models import Model, load_model, list_models
from pycollimator.projects import get_project
from pycollimator.simulations import run_simulation, linearize

__all__ = [
    "Api",
    "Log",
    "Model",
    "load_model",
    "list_models",
    "get_project",
    "get_project_url",
    "linearize",
    "run_simulation",
    "set_auth_token",
]
